<?php

/**
 * Object-oriented CDDB Server components 
 * 
 * Alpha-quality CDDB server capable of serving CDDB requests over a variety of 
 * CDDB protocol options (HTTP only for now, more coming soon) and using a 
 * variety of data sources. 
 * 
 * @see Net_CDDB_Client
 * 
 * @author Keith Palmer <Keith@UglySlug.com>
 * @category Net
 * @package Net_CDDB
 * @license http://www.opensource.org/licenses/bsd-license.php BSD License
 */

/**
 * Need the constants from this file...
 */
require_once 'Net/CDDB.php';

/**
 * CDDB Server (ALPHA QUALITY, *USE AT YOUR OWN RISK!*)
 * 
 * *** WARNING *** 
 * Use at your own risk, it doesn't work very well yet!
 * 
 * Current features: 
 * 	- CDDB protocol level 5
 * 	- Multiple data sources: tell the server to read from a local filesystem, a remote HTTP CDDB server, a remote CDDBP server, a database (coming soon) etc.
 * 	- Supports about half of the CDDB commands  
 * 
 * Known issues: 
 * 	- Extremelly lightly tested (use at your own risk!)
 * 	- Command: 'motd' reports and incorrect last modified date
 * 	- Lack of any error reporting...
 * 	- Lack of any login/hello command support
 * 	- Only supports protocol level 5
 * 	- Only provides HTTP request support (no CDDBP or SMTP... yet)
 * 
 * @todo Probably have the Net_CDDB_Request_* classes have three methods: getHello(), getProto(), and getCmd() or something
 * 
 * @package Net_CDDB
 */
class Net_CDDB_Server extends Net_CDDB
{
	/**
	 * Net_CDDB_Request_* object instance to use for listening for requests
	 * 
	 * @access protected
	 * @var object
	 */
	var $_request;
	
	/**
	 * Net_CDDB_Protocol_* object instance to use for data access
	 * 
	 * @access protected
	 * @var object
	 */
	var $_protocol;
	
	/*var $_connect_logging;
	
	var $_use_auth_file;
	var $_auth_file;
	var $_use_auth_mysql;
	var $_auth_method;
	*/
	
	/**
	 * Constructor (PHP v4.x)
	 * 
	 * @access public
	 * @see Net_CDDB_Server::__construct()
	 */
	function Net_CDDB_Server($request_dsn, $protocol_dsn, $options = array())
	{
		$this->__construct($request_dsn, $protocol_dsn, $options);
	}
	
	/**
	 * Constructor (PHP v5.x)
	 * 
	 * @access public
	 * 
	 * @param string $request_dsn
	 * @param string $protocol_dsn
	 * @param array $options
	 */
	function __construct($request_dsn, $protocol_dsn, $options = array())
	{
		$this->_protocol = $this->_createProtocol($protocol_dsn, $options);
		$this->_request = $this->_createRequest($request_dsn, $options);
	}
	
	/**
	 * Create a Net_CDDB_Request object instance
	 * 
	 * @access protected
	 * 
	 * @param string $dsn
	 * @param array $options
	 * @return Net_CDDB_Request Returns an instance of a subclass of Net_CDDB_Request (i.e: Net_CDDB_Request_HTTP instance)
	 */
	function _createRequest($dsn, $options)
	{
		$parse = parse_url($dsn);
		
		$file = ucfirst(strtolower($parse['scheme']));
		$class = 'Net_CDDB_Request_' . $file;
		
		/**
		 * Require the file the protocol class is stored in
		 */
		include_once 'Net/CDDB/Request/' . $file . '.php';
		
		if (class_exists($class)) {
			return new $class($dsn, $options);
		} else {
			return PEAR::raiseError('Could not find request file for: ' . $file);
		}
	}
	
	/**
	 * Start the CDDB server
	 * 
	 * @access public
	 * @todo Listen forever support (daemon mode)
	 * 
	 * @return boolean 
	 */
	function start()
	{
		if (PEAR::isError($this->_request)) {
			return $this->_request;
		}
		
		if (PEAR::isError($this->_protocol)) {
			return $this->_request;
		}
		
		// Implemented commands
		$impl_cmds = array( 
			'cddb lscat' 	=> '_cddbLscat', 
			'cddb hello' 	=> '_cddbHello', 
			'cddb query' 	=> '_cddbQuery',
			'proto' 		=> '_proto',
			'cddb read' 	=> '_cddbRead', 
			'motd' 			=> '_motd', 
			'ver' 			=> '_ver', 
			);
		
		if ($this->_request->mode() == NET_CDDB_REQUEST_ONCE) { // Listen just once (HTTP for instance)
			
			// Read data from request listener
			$commands = $this->_request->listen();
			
			if (!count($commands))
			{
				$this->_request->respond(NET_CDDB_RESPONSE_ERROR_SYNTAX, 'Command syntax error: incorrect number of arguments.');
				return true;
			}
			
			foreach ($commands as $cmd)
			{
				// Parse command from command parameters
				$tmp = explode(' ', $cmd);
				if (current($tmp) == 'cddb' and count($tmp) >= 2) {
					$cmd = 'cddb ' . next($tmp);
					$query = implode(' ', array_slice($tmp, 2));
				} else {
					$cmd = current($tmp);
					$query = implode(' ', array_slice($tmp, 1));
				}
				
				if (!strlen($cmd)) {
					// Empty command...?
					$this->_request->respond(NET_CDDB_RESPONSE_ERROR_EMPTY, 'Empty command input.');
					return true;
				} else if (isset($impl_cmds[$cmd]) and method_exists($this, $impl_cmds[$cmd])) {
					// Call the appropriate command handler (the command handler will respond)
					$this->{$impl_cmds[$cmd]}($cmd, $query);
				} else {
					// Respond indicating the command could not be found
					$this->_request->respond(NET_CDDB_RESPONSE_ERROR_UNRECOGNIZED, 'Unrecognized command.'); // Unrecognized command...?
					return true;
				}
			}
			
			return true;
			
		} else {
			// listen forever, not implemented yet
		}
	}
	
	/**
	 * Dummy placeholder for the 'cddb proto' command
	 * 
	 * @access protected
	 * 
	 * @return void
	 */
	function _proto()
	{
		return;
	}
	
	/**
	 * Dummy placeholder for 'cddb hello' command
	 * 
	 * @access protected
	 * 
	 * @return void
	 */
	function _cddbHello()
	{
		return;	
	}
	
	/**
	 * Respond to the 'cddb lscat' CDDB command (list categories)
	 * 
	 * @access protected
	 * 
	 * @param string $cmd
	 * @param string $query
	 * @return void
	 */
	function _cddbLscat($cmd, $query)
	{
		$headers = array(
			NET_CDDB_RESPONSE_OK_FOLLOWS => 'OK, category list follows (until terminating `.\')',
			);
		
		$this->_protocol->send('cddb lscat');
		$resp = $this->_protocol->recieve();
		
		$status = $this->_protocol->status();
		
		switch ($status)
		{
			case NET_CDDB_RESPONSE_OK_FOLLOWS:
				$this->_request->respond($status, $headers[$status], $resp, true);
				break;
			default:
				$this->_request->respond(NET_CDDB_RESPONSE_SERVER_ERROR, 'Internal server error.');
				break;
		}
	}
	
	/**
	 * Respond to the 'cddb read ...' CDDB command
	 * 
	 * @access protected
	 * 
	 * @param string $cmd
	 * @param string $query
	 * @return void
	 */
	function _cddbRead($cmd, $query)
	{
		$headers = array(
			NET_CDDB_RESPONSE_OK_FOLLOWS => $query . ' CD database entry follows (until terminating `.\')', 
			);
		
		$this->_protocol->send($cmd . ' ' . $query);
		$resp = $this->_protocol->recieve();
		
		$status = $this->_protocol->status();
		
		switch ($status)
		{
			case NET_CDDB_RESPONSE_OK_FOLLOWS:
				$this->_request->respond($status, $headers[$status], $resp, true);
				break;
			default:
				$this->_request->respond(NET_CDDB_RESPONSE_SERVER_ERROR, 'Internal server error.');
				break;
		}
	}
	
	/**
	 * Respond to the 'cddb query ...' CDDB command
	 * 
	 * @access protected
	 * 
	 * @param string $cmd
	 * @param string $query
	 * @return void
	 */
	function _cddbQuery($cmd, $query)
	{
		$this->_protocol->send($cmd . ' ' . $query);
		$resp = $this->_protocol->recieve();
		
		$status = $this->_protocol->status();
		
		switch ($status)
		{
			case NET_CDDB_RESPONSE_OK:
				$this->_request->respond($status, $resp);
				break;
			default:
				$this->_request->respond(NET_CDDB_RESPONSE_SERVER_ERROR, 'Internal server error.');
				break;
		}
	}
	
	/**
	 * Respond to the CDDB 'motd' command
	 * 
	 * @access protected
	 * @todo Find out the motd modified time somehow...?
	 * 
	 * @param string $cmd
	 * @param string $query
	 * @return void 
	 */
	function _motd($cmd, $query)
	{
		$headers = array(
			NET_CDDB_RESPONSE_OK_FOLLOWS => 'Last modified: ' . date('m/d/Y H:i:s') . ' MOTD follows (until terminating `.\')', 
			);
		
		$this->_protocol->send('motd');
		$resp = $this->_protocol->recieve();
		
		$status = $this->_protocol->status();
		
		switch ($status)
		{
			case NET_CDDB_RESPONSE_OK_FOLLOWS:
				$this->_request->respond($status, $headers[$status], $resp, true);
				break;
			default:
				$this->_request->respond(NET_CDDB_RESPONSE_SERVER_ERROR, 'Internal server error.');
				break;
		}
	}
	
	/**
	 * Respond to the CDDB 'ver' command
	 * 
	 * @access protected
	 * 
	 * @param string $cmd
	 * @param string $query
	 * @return void
	 */
	function _ver($cmd, $query)
	{
		$this->_protocol->send('ver');
		$resp = $this->_protocol->recieve();
		
		$status = $this->_protocol->status();
		
		/*
		 * If the protocol we're using to get the data is a remote protocol 
		 * (i.e.: We're reading this data from a remote CDDB server via 
		 * HTTP/CDDBP instead of using a local copy of the CDDB db) then we'll 
		 * modify the version information to reflect the tunnelling of the 
		 * response through the Net_CDDB/PHP server. 
		 */ 
		if ($this->_protocol->remote()) {
			$resp .= ' (Tunnelled through PHP/PEAR/' . get_class($this) . ' v' . NET_CDDB_VERSION . ')';
		}
		
		switch ($status)
		{
			case NET_CDDB_RESPONSE_OK:
				$this->_request->respond($status, $resp);
				break;
			default:
				$this->_request->respond(NET_CDDB_RESPONSE_SERVER_ERROR, 'Internal server error.');
				break;
		}
	}
}

?>
